@@ -26,7 +26,7 @@ import Paths_alex            ( version, getDataDir )
 import Control.Exception     ( bracketOnError )
 import Control.Monad         ( when, liftM )
 import Data.Char             ( chr )
-import Data.List             ( isSuffixOf, nub )
+import Data.List             ( intercalate, isSuffixOf, nub )
 import Data.Map              ( Map )
 import Data.Version          ( showVersion )
 import System.Console.GetOpt ( getOpt, usageInfo, ArgOrder(..), OptDescr(..), ArgDescr(..) )
@@ -364,56 +364,82 @@ injectCode (Just (AlexPn _ ln _,code)) filename hdl = do
   hPutStrLn hdl code
 
 optsToInject :: Target -> [CLIFlags] -> [String]
-optsToInject target _ = optNoWarnings : "{-# LANGUAGE CPP #-}" : case target of
-  GhcTarget -> ["{-# LANGUAGE MagicHash #-}"]
-  _         -> []
+optsToInject target _ = concat
+  [ optNoWarnings
+  , [ "{-# LANGUAGE CPP #-}" ]
+  , [ "{-# LANGUAGE MagicHash #-}" | target == GhcTarget ]
+  ]
 
-optNoWarnings :: String
-optNoWarnings = "{-# OPTIONS_GHC -fno-warn-unused-binds -fno-warn-missing-signatures #-}"
+-- List here all harmless warnings caused by Alex-generated code.
+--
+-- These will be suppressed so that they are not printed
+-- when users turn on @-Wall@ in their lexer project.
+--
+optNoWarnings :: [String]
+optNoWarnings =
+  map (("{-# OPTIONS_GHC -fno-warn-" ++) . (++ " #-}"))
+    [ "missing-signatures"
+    , "tabs"
+    , "unused-binds"
+    , "unused-imports"
+    ]
 
 importsToInject :: Target -> [CLIFlags] -> String
-importsToInject _ cli = always_imports ++ debug_imports ++ glaexts_import
+importsToInject _ cli = unlines $
+  always_imports ++ debug_imports ++ glaexts_import
   where
         glaexts_import | OptGhcTarget `elem` cli    = import_glaexts
-                       | otherwise                  = ""
+                       | otherwise                  = []
 
         debug_imports  | OptDebugParser `elem` cli = import_debug
-                       | otherwise                 = ""
+                       | otherwise                 = []
 
--- CPP is turned on for -fglasogw-exts, so we can use conditional
--- compilation.  We need to #include "config.h" to get hold of
+-- We need to #include "ghcconfig.h" to get hold of
 -- WORDS_BIGENDIAN (see AlexTemplate.hs).
 
-always_imports :: String
-always_imports = "#if __GLASGOW_HASKELL__ >= 603\n" ++
-                 "#include \"ghcconfig.h\"\n" ++
-                 "#elif defined(__GLASGOW_HASKELL__)\n" ++
-                 "#include \"config.h\"\n" ++
-                 "#endif\n" ++
-                 "#if __GLASGOW_HASKELL__ >= 503\n" ++
-                 "import Data.Array\n" ++
-                 "#else\n" ++
-                 "import Array\n" ++
-                 "#endif\n"
+always_imports :: [String]
+always_imports =
+  [ "#include \"ghcconfig.h\""
+  , "import qualified Data.Array"
+  ]
 
-import_glaexts :: String
-import_glaexts = "#if __GLASGOW_HASKELL__ >= 503\n" ++
-                 "import Data.Array.Base (unsafeAt)\n" ++
-                 "import GHC.Exts\n" ++
-                 "#else\n" ++
-                 "import GlaExts\n" ++
-                 "#endif\n"
+import_glaexts :: [String]
+import_glaexts =
+  [ "import Data.Array.Base (unsafeAt)"
+  , "import GHC.Exts (" ++ imports ++ ")"
+  , "import qualified GHC.Exts"
+  ]
+  where
+    -- We can import anything mentioning # safely,
+    -- assuming the user code does not make use of
+    -- MagicHash.
+    imports = intercalate ","
+      [ "Addr#"
+      , "Int#"
+      , "Int(I#)"
+      , "(*#)"
+      , "(+#)"
+      , "(-#)"
+      , "(==#)"
+      , "(>=#)"
+      , "indexCharOffAddr#"
+      , "indexInt16OffAddr#"
+      , "indexInt32OffAddr#"
+      , "int2Word#"
+      , "narrow16Int#"
+      , "narrow32Int#"
+      , "negateInt#"
+      , "or#"
+      , "ord#"
+      , "uncheckedShiftL#"
+      , "word2Int#"
+      ]
 
-import_debug :: String
-import_debug   = "#if __GLASGOW_HASKELL__ >= 503\n" ++
-                 "import Data.Char (chr)\n" ++
-                 "import System.IO\n" ++
-                 "import System.IO.Unsafe\n" ++
-                 "import Debug.Trace\n" ++
-                 "#else\n" ++
-                 "import IO\n" ++
-                 "import IOExts\n" ++
-                 "#endif\n"
+import_debug :: [String]
+import_debug =
+  [ "import Data.Char (chr)"
+  , "import qualified Debug.Trace"
+  ]
 
 templateDir :: IO FilePath -> [CLIFlags] -> IO FilePath
 templateDir def cli
